/*
 * This file is part of the LibreOffice project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#include "RenderScene.h"
#include <glm/gtx/transform.hpp>
#include <glm/gtc/matrix_inverse.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/quaternion.hpp>

RenderPrimitive::RenderPrimitive()
    : mVerterCount(0), mIndicesCount(0)
    , pMatrix(0), pProject(0), pColor(0), pMaterial(0)
    , mShaderVertexId(-1), mShaderNormalId(-1), mShaderCoord0Id(-1)
    , mShaderJointId(-1), mShaderWeightId(-1)
{
}

RenderPrimitive::~RenderPrimitive()
{
    glDeleteBuffers(1, &mVertexBuffer);
    glDeleteBuffers(1, &mNormalBuffer);
    glDeleteBuffers(1, &mTexCoordBuffer);
    glDeleteBuffers(1, &mIndicesBuffer);

    delete pProject;
    delete pColor;
}

Material* RenderPrimitive::getMaterial()
{
    return this->pMaterial;
}

void RenderPrimitive::setMaterial(Material* pMaterial)
{
    this->pMaterial = pMaterial;
}

unsigned int RenderPrimitive::getVerterCount()
{
    return this->mVerterCount;
}

void RenderPrimitive::setVerterCount(unsigned int count)
{
    this->mVerterCount = count;
}

unsigned int RenderPrimitive::getIndicesCount()
{
    return this->mIndicesCount;
}

void RenderPrimitive::setIndicesCount(unsigned int count)
{
    this->mIndicesCount = count;
}

DataType_E RenderPrimitive::getIndicesDataType()
{
    return this->mIndicesDataType;
}

void RenderPrimitive::setIndicesDataType(DataType_E type)
{
    this->mIndicesDataType = type;
}

int RenderPrimitive::getShaderVertexId()
{
    return this->mShaderVertexId;
}

void RenderPrimitive::setShaderVertexId(int id)
{
    this->mShaderVertexId = id;
}

int RenderPrimitive::getShaderNormalId()
{
    return this->mShaderNormalId;
}

void RenderPrimitive::setShaderNormalId(int id)
{
    this->mShaderNormalId = id;
}

int RenderPrimitive::getShaderCoord0Id()
{
    return this->mShaderCoord0Id;
}

void RenderPrimitive::setShaderCoord0Id(int id)
{
    this->mShaderCoord0Id = id;
}

int RenderPrimitive::getShaderWeightId()
{
    return this->mShaderWeightId;
}

void RenderPrimitive::setShaderWeightId(int id)
{
    this->mShaderWeightId = id;
}

int RenderPrimitive::getShaderJointId()
{
    return this->mShaderJointId;
}

void RenderPrimitive::setShaderJointId(int id)
{
    this->mShaderJointId = id;
}

unsigned int RenderPrimitive::getVertexBuffer()
{
    return this->mVertexBuffer;
}

void RenderPrimitive::setVertexBuffer(unsigned int bufferId)
{
    this->mVertexBuffer = bufferId;
}

unsigned int RenderPrimitive::getNormalBuffer()
{
    return this->mNormalBuffer;
}

void RenderPrimitive::setNormalBuffer(unsigned int bufferId)
{
    this->mNormalBuffer = bufferId;
}

unsigned int RenderPrimitive::getTexCoordBuffer()
{
    return this->mTexCoordBuffer;
}

void RenderPrimitive::setTexCoordBuffer(unsigned int bufferId)
{
    this->mTexCoordBuffer = bufferId;
}

unsigned int RenderPrimitive::getJointBuffer()
{
    return this->mJointBuffer;
}

void RenderPrimitive::setJointBuffer(unsigned int bufferId)
{
    this->mJointBuffer = bufferId;
}

unsigned int RenderPrimitive::getWeightBuffer()
{
    return this->mWeightBuffer;
}

void RenderPrimitive::setWeightBuffer(unsigned int bufferId)
{
    this->mWeightBuffer = bufferId;
}

unsigned int RenderPrimitive::getIndicesBuffer()
{
    return this->mIndicesBuffer;
}

void RenderPrimitive::setIndicesBuffer(unsigned int bufferId)
{
    this->mIndicesBuffer = bufferId;
}

void RenderPrimitive::setProject(glm::mat4* project)
{
    this->pProject = project;
}

glm::mat4* RenderPrimitive::getProject()
{
    return this->pProject;
}

void RenderPrimitive::setColor(glm::vec3* color)
{
    this->pColor = color;
}

glm::vec3* RenderPrimitive::getColor()
{
    return this->pColor;
}

void RenderPrimitive::setMatrix(glm::mat4* matrix)
{
    this->pMatrix = matrix;
}

glm::mat4* RenderPrimitive::getMatrix()
{
    return this->pMatrix;
}

RenderScene::RenderScene()
	: pScene (0), pLight (0), pCamera(0)
    , mCurrentTime(0), mLastPlaying(0), mDuration(0)
    , mAnimationPlay(false), mAnimationLoop(true)
{
#if ENABLE_FPS
    pFPSCounter = NULL;
#endif
}

RenderScene::~RenderScene()
{
#if ENABLE_FPS
    delete pFPSCounter;
#endif
    if (pScene->getLight())
        delete pLight;
    delete pCamera;
    RenderPrimitive* pRenderPrimitive;
    unsigned int size = mPrimitiveVec.size();
    for (unsigned int i = 0; i < size; ++i)
    {
        pRenderPrimitive = mPrimitiveVec[i];
        delete pRenderPrimitive;
    }
    mPrimitiveVec.clear();
    delete pScene;
}

Scene* RenderScene::loadScene(const std::string& directory, 
    const std::string& file)
{
    Scene* pScene = new Scene();
    loadJson.setJsonInfo(directory, file);
    if (!loadJson.parseJsonFile())
    {
        return 0;
    }

    loadJson.parseScene(pScene);
    return pScene;
}

Scene* RenderScene::loadScene(glTFHandle* handle)
{ 
    Scene* pScene = new Scene();
    pScene->setGltfHandle(handle);
    if (!loadJson.parseScene(pScene))
        return 0;
    return pScene;
}

glTFHandle* RenderScene::initScene(const glTFFile* jsonfile)
{
    if(0 == jsonfile)
        return 0;
    string tmpStr(jsonfile->buffer, jsonfile->size);
    loadJson.setJsonInfo(tmpStr);
    if (!loadJson.parseJsonFile())
        return 0;
    string tmpFileName(jsonfile->filename);
    return loadJson.getFileNameInJson(tmpFileName);
}

unsigned int RenderScene::bindAttribute(const Attribute* pAttr)
{
    unsigned int bufferId;
    glGenBuffers(1, &bufferId);
    glBindBuffer(GL_ARRAY_BUFFER, bufferId);
    glBufferData(GL_ARRAY_BUFFER, pAttr->getDataCount() * pAttr->getByteStride(),
            pAttr->getAttributeData(), GL_STATIC_DRAW);
    return bufferId;
}

unsigned int RenderScene::bindIndices(const Attribute* pAttr)
{
    unsigned int bufferId;
    glGenBuffers(1, &bufferId);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bufferId);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 
                pAttr->getDataCount() * pAttr->getByteStride(),
				pAttr->getAttributeData(), GL_STATIC_DRAW);
    return bufferId;
}

void RenderScene::constructMesh(const std::string meshName, glm::mat4* pmatrix)
{
    const Mesh* pMesh = pScene->findMesh(meshName);
    unsigned int size = pMesh->getPrimitiveVecSize();
    for (unsigned int i = 0; i < size; ++i)
        constructPrimitive(pMesh->getPrimitiveVec(i), pmatrix);
}

void RenderScene::constructPrimitive(const Primitives* pPrimitive, 
    glm::mat4* pmatrix)
{
    Material* pMaterial;
    pMaterial = pScene->findMaterial(pPrimitive->getMaterialIndex());
    if (0 == pMaterial)
        return;
    RenderPrimitive* pRenderPrimitive = new RenderPrimitive();

    getAttributeShaderId(pMaterial, pRenderPrimitive);
    bindAttributeBuffer(pPrimitive, pRenderPrimitive);
    pRenderPrimitive->setMaterial(pMaterial);
    glm::mat4* pProject = calcProjectionMatrix();
    pRenderPrimitive->setProject(pProject);
    glm::vec3* pColor = calcLightColor();
    pRenderPrimitive->setColor(pColor);
    pRenderPrimitive->setMatrix(pmatrix);
    setPrimitiveVec(pRenderPrimitive);
}

void RenderScene::getAttributeShaderId(Material* pMaterial,
    RenderPrimitive* pRP)
{
    Technique* pTechnique = pScene->findTechnique(pMaterial->getTechniqueId());
    if (0 == pTechnique)
    {
        return;
    }

    TechAttribute* pAttribute = pTechnique->getTechAttribute("a_position");
    if (0 != pAttribute)
    {
        pRP->setShaderVertexId(glGetAttribLocation(
            pTechnique->getProgramId(),
            pAttribute->getAttributeName().c_str()));
    }

    pAttribute = pTechnique->getTechAttribute("a_normal");
    if (0 != pAttribute)
    {
        pRP->setShaderNormalId(glGetAttribLocation(
            pTechnique->getProgramId(),
            pAttribute->getAttributeName().c_str()));
    }

    pAttribute = pTechnique->getTechAttribute("a_texcoord0");
    if (0 != pAttribute)
    {
        pRP->setShaderCoord0Id(glGetAttribLocation(
            pTechnique->getProgramId(),
            pAttribute->getAttributeName().c_str()));
    }

    pAttribute = pTechnique->getTechAttribute("a_weight");
    if (0 != pAttribute)
    {
        pRP->setShaderWeightId(glGetAttribLocation(
            pTechnique->getProgramId(),
            pAttribute->getAttributeName().c_str()));
    }

    pAttribute = pTechnique->getTechAttribute("a_joint");
    if (0 != pAttribute)
    {
        pRP->setShaderJointId(glGetAttribLocation(
                pTechnique->getProgramId(),
                pAttribute->getAttributeName().c_str()));
    }
}

void RenderScene::bindAttributeBuffer(const Primitives* pPrimitive,
    RenderPrimitive* pRP)
{
    const Attribute* pAttr;
    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("POSITION"));
    if (0 != pAttr)
    {
        pRP->setVertexBuffer(bindAttribute(pAttr));
        pRP->setVerterCount(pAttr->getDataCount());
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("NORMAL"));
    if (0 != pAttr)
    {
        pRP->setNormalBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("TEXCOORD_0"));
    if (0 != pAttr)
    {
        float* pData = (float*)pAttr->getAttributeData();
        unsigned int size = pAttr->getDataCount();
        for (unsigned int i = 0; i < size; i++)
        {
            pData[i*2+1] = 1 - pData[i*2+1];
        }
        pRP->setTexCoordBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("JOINT"));
    if (0 != pAttr)
    {
        pRP->setJointBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("WEIGHT"));
    if (0 != pAttr)
    {
        pRP->setWeightBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getIndicesIndex());
    if (0 != pAttr)
    {
        pRP->setIndicesBuffer(bindIndices(pAttr));
        pRP->setIndicesCount(pAttr->getDataCount());
        pRP->setIndicesDataType(pAttr->getDataType());
    }
}

void RenderScene::setPrimitiveVec(RenderPrimitive* pPrim)
{
    (this->mPrimitiveVec).push_back(pPrim);
}

RenderPrimitive* RenderScene::getPrimitiveVec(unsigned int index)
{
    if (this->mPrimitiveVec.size() <= index)
        return 0;
    return this->mPrimitiveVec[index];
}

unsigned int RenderScene::getPrimitiveVecSize()
{
    return this->mPrimitiveVec.size();
}

glm::mat4* RenderScene::calcProjectionMatrix()
{
    const Node* pParentNode = pScene->getRootNode();
    glm::mat4 projection;
    if (!pParentNode->getFlagCamera())
    {
        projection = glm::perspective(37.8492f, 1.5f, 1.0f, 10000.0f);
    }
    else
    {
        Camera* cCamera  = pScene->findCamera(pScene->getCameraIndex());
        projection = glm::perspective(cCamera->getYFov(),
            cCamera->getAspectRatio(),cCamera->getNear(), cCamera->getFar());
    }
    glm::mat4* pProject = new glm::mat4;

    memcpy(pProject, glm::value_ptr(projection), sizeof(glm::mat4));
    return pProject;
}

glm::vec3* RenderScene::calcLightColor()
{
    glm::vec3 *pLightColor = new glm::vec3(pLight->getColor());
    return pLightColor;
}

void RenderScene::initOpengl()
{
    glClearColor(0.0f, 0.0f, 1.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glDepthFunc(GL_LESS);
}

void RenderScene::createDefaultLight()
{
    if (pScene->getLight())
    {
        pLight = new Light();
        light0Transfrom = new glm::mat4;
        memset((void *)light0Transfrom, 0, sizeof(glm::mat4));
    }
    else
    {
        pLight = pScene->findLight(pScene->GetLightIndex());
        light0Transfrom = pScene->findLightTransformation();
    }
    glLightfv(GL_LIGHT0, GL_AMBIENT, pLight->getColorAmbient());
    glLightfv(GL_LIGHT0, GL_DIFFUSE, pLight->getColorDiffuse());
    glLightfv(GL_LIGHT0, GL_SPECULAR, pLight->getColorSpecular());
    glLightf( GL_LIGHT0, GL_SPOT_CUTOFF, 45.0 );
}

void RenderScene::createDefaultCamera()
{
    glm::vec3 vDelta = pScene->getVertexMax() - pScene->getVertexMin();
    glm::vec3 vHalf =  pScene->getVertexMin() + (vDelta / 2.0f);

    float fPos = sqrt(vDelta.x*vDelta.x+vDelta.y*vDelta.y+vDelta.z*vDelta.z);
    if (fPos >= 3000)
    {
        pCamera = new CPhysicalCamera(glm::vec3(0.0f, -fPos/5, fPos*0.8f),
            vHalf,
            glm::vec3(0.0f, 1.0f, 0.0f));
    } 
    else
    {
        pCamera = new CPhysicalCamera(glm::vec3(fPos, -fPos, 2*fPos),
            vHalf,
            glm::vec3(.0f, 1.0f, 0.0f));
    }
    pCamera->flength = fPos;
    pCamera->SetRotationAngle();
    if (sCamera==NULL)
        sCamera = pCamera;   
    else if (sCamera->flength <=pCamera->flength)
        sCamera = pCamera;
}

void RenderScene::initRender(glTFHandle* handle)
{ 
    initOpengl();
    pScene = loadScene(handle);
    if (0 == pScene)
        return;

    createDefaultLight();
    createDefaultCamera();
    const Node* pNode;
    unsigned int nSize = pScene->getNodeSize();
    for (unsigned int i = 0; i < nSize; ++i)
    {
        pNode = pScene->getNode(i);
        if (!pNode->hasMesh())
            continue;
        unsigned int mSize = pNode->getMeshIndexSize();
        for (unsigned int j = 0; j < mSize; ++j)
            constructMesh(pNode->getMeshIndex(j), pNode->getMatrix());
    }

    Animation* pAnimation;
    if (0 != pScene->getAnimationCount())
    {
        pAnimation = pScene->getAnimation(0);
        this->mDuration = pAnimation->getDuration();
    }

    if (pScene->getSkin())
        updateBoneMatrix(pScene->getRootBone(), glm::mat4(1.0f));

#if ENABLE_FPS
    pFPSCounter =new FPSCounter();
    pFPSCounter->loadFPSShader(&mShaderProgram);
#endif
}

void RenderScene::renderPrimitive(RenderPrimitive* pPrimitive)
{
    Material* pMaterial = pPrimitive->getMaterial();
    Technique* pTechnique = 
        pScene->findTechnique(pMaterial->getTechniqueId());
    if (!pTechnique->getProgramState())
        return;
    pTechnique->useTechnique();
    unsigned int progId = pTechnique->getProgramId();
    glm::mat4* pMatrix = pPrimitive->getMatrix();
    glm::mat4 modelViewMatrix = sCamera->getModelViewMatrix(pMatrix);
    mShaderProgram.setUniform(progId, "u_modelViewMatrix", modelViewMatrix);

    glm::mat3 normalMatrix(*pMatrix);
    glm::mat3 NormalInverseTranspose = glm::mat3(sCamera->Look()) *
        (glm::transpose(glm::inverse(normalMatrix)));
    mShaderProgram.setUniform(progId, "u_normalMatrix", NormalInverseTranspose);
    mShaderProgram.setUniform(progId, "u_projectionMatrix", 
        pPrimitive->getProject());
    mShaderProgram.setUniform(progId, "u_light0Color", pPrimitive->getColor());
    mShaderProgram.setUniform(progId, "u_light0Transform", light0Transfrom);

    if (pScene->hasSkin())
    {
        if (!this->mAnimationLoop &&
            (fmod(mCurrentTime, mDuration) < fmod(mLastPlaying, mDuration)))
            this->mAnimationPlay = false;
        if (this->mAnimationPlay)
        {
            updateBoneMatrix(pScene->getRootBone(), glm::mat4(1.0f));
        }
        Skin* pSkin = pScene->getSkin();
        glm::mat4* pBindMatrix = pSkin->getBindMatrix();
        glm::mat4* pTempMatrix = new glm::mat4[pSkin->getBindMatrixCount()];
        memcpy(pTempMatrix, pBindMatrix, sizeof(glm::mat4)*pSkin->getBindMatrixCount());

        unsigned int jointCount = pSkin->getBoneIdSize();
        for (unsigned int i = 0; i < jointCount; i++)
        {
            Bone* pBone = pScene->findBone(pSkin->getBoneId(i));
            pTempMatrix[i] = pBone->getGlobleMatrix() * pTempMatrix[i];
        }

        int iLoc = glGetUniformLocation(progId, "u_jointMat");
        glUniformMatrix4fv(iLoc, pSkin->getBindMatrixCount(), false, (GLfloat*)pTempMatrix);
        delete[] pTempMatrix;
    }

    int textureId = 0;
    Texture* oneTexture;
    unsigned int size = pMaterial->getMaterialProperSize();
    for (unsigned int i = 0; i < size; ++i)
    {
        const MaterialProperty* pProper = pMaterial->getMaterialProper(i);
        std::string properName = "u_" + pProper->getPropertyName();
        switch (pProper->getDataType())
        {
            case DataType_FLOAT:
                mShaderProgram.setUniform(progId, properName.c_str(),
                        (float*)(pProper->getPropertyData()));
                break;
            case DataType_FLOAT_VEC2:
                mShaderProgram.setUniform(progId, properName.c_str(),
                        (glm::vec2*)(pProper->getPropertyData()));
                break;
            case DataType_FLOAT_VEC3:
                mShaderProgram.setUniform(progId, properName.c_str(),
                        (glm::vec3*)(pProper->getPropertyData()));
                break;
            case DataType_FLOAT_VEC4:
                mShaderProgram.setUniform(progId, properName.c_str(),
                        (glm::vec4*)(pProper->getPropertyData()));
                break;
            case DataType_FLOAT_MAT3:
                mShaderProgram.setUniform(progId, properName.c_str(),
                        (glm::mat3*)(pProper->getPropertyData()));
                break;
            case DataType_FLOAT_MAT4:
                mShaderProgram.setUniform(progId, properName.c_str(),
                        (glm::mat4*)(pProper->getPropertyData()));
                break;
            case DataType_SAMPLER_2D:
                mShaderProgram.setUniform(progId, properName.c_str(), textureId);
                oneTexture = pScene->findTexture(pProper->getImagePath());
                oneTexture->bindTexture(textureId);
                textureId++;
                break;
            default:
                break;
        }
    }
    upLoadAttribute(pPrimitive->getShaderVertexId(),
            pPrimitive->getVertexBuffer(), 3);
    if (-1 != pPrimitive->getShaderNormalId())
    {   
        upLoadAttribute(pPrimitive->getShaderNormalId(), 
            pPrimitive->getNormalBuffer(), 3);
    }
    if (-1 != pPrimitive->getShaderCoord0Id())
    {
        upLoadAttribute(pPrimitive->getShaderCoord0Id(),
            pPrimitive->getTexCoordBuffer(), 2);
    }
    if (-1 != pPrimitive->getShaderJointId())
    {
        upLoadAttribute(pPrimitive->getShaderJointId(),
            pPrimitive->getJointBuffer(), 4);
    }
    if (-1 != pPrimitive->getShaderWeightId())
    {
        upLoadAttribute(pPrimitive->getShaderWeightId(),
            pPrimitive->getWeightBuffer(), 4);
    }

    setStates(pTechnique);
#if ENABLE_FPS
    pFPSCounter->gpuTimeStampStart();
    drawTriangle(pPrimitive);
    pFPSCounter->gpuTimeStampEnd();
    pFPSCounter->gpuTimeAccumulation();
#else
    drawTriangle(pPrimitive);
#endif
    glUseProgram(0);
    glDisableVertexAttribArray(pPrimitive->getShaderVertexId());
    if (-1 != pPrimitive->getShaderNormalId())
        glDisableVertexAttribArray(pPrimitive->getShaderNormalId());
    if (-1 != pPrimitive->getShaderCoord0Id())
        glDisableVertexAttribArray(pPrimitive->getShaderCoord0Id());
    if (-1 != pPrimitive->getShaderJointId())
        glDisableVertexAttribArray(pPrimitive->getShaderJointId());
    if (-1 != pPrimitive->getShaderWeightId())
        glDisableVertexAttribArray(pPrimitive->getShaderWeightId());
}

void RenderScene::upLoadAttribute(unsigned int shaderId, unsigned int bufferId, 
    int size)
{
    glEnableVertexAttribArray(shaderId);
    glBindBuffer(GL_ARRAY_BUFFER, bufferId);
    glVertexAttribPointer(shaderId, size, GL_FLOAT, GL_FALSE, 0, (void*)0);
    return;
}

void RenderScene::drawTriangle(RenderPrimitive* pPrimitive)
{
    if (pPrimitive->getIndicesCount() > 0)
    {
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, pPrimitive->getIndicesBuffer());
        glDrawElements(GL_TRIANGLES, pPrimitive->getIndicesCount(), 
                pPrimitive->getIndicesDataType(), 0);
    }
    else
    {
        glDrawArrays(GL_TRIANGLES, 0, pPrimitive->getVerterCount());
    }
}

void RenderScene::prepareRender(void* lpParam)
{
#if ENABLE_FPS
  pFPSCounter->timeStamp();
  pFPSCounter->gpuAccumulationTimeclear();
#endif
  viewport = *((glTFViewport*)lpParam);
  glViewport(viewport.x, viewport.y, viewport.width, viewport.height);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  return;
}

void RenderScene::render()
{
    mCurrentTime += clock() / double(CLOCKS_PER_SEC) - mLastPlaying;
    unsigned int pSize = getPrimitiveVecSize();
    for (unsigned int i = 0; i < pSize; ++i)
    {
        renderPrimitive(getPrimitiveVec(i));
    }
    mLastPlaying = mCurrentTime;
}

void RenderScene::completeRender()
{
#if ENABLE_FPS
  pFPSCounter->printFPS(&viewport);
#endif
  SwapBuffers(wglGetCurrentDC());
  return;
}

void RenderScene::releaseRender(void* lpParam)
{
    loadJson.releaseFileName();    
    return;
}

void RenderScene::updateJointInfo(Bone* pBone)
{ 
    Animation* pAnimation;
    AnimChannel* pAnimChannel;

    double time = fmod(this->mCurrentTime, this->mDuration);
    unsigned int aCount = pScene->getAnimationCount();
    for (unsigned int i = 0; i < aCount; i++)
    {
        pAnimation = pScene->getAnimation(i);
        //time = fmod(time, pAnimation->getDuration());   
        pAnimChannel = pAnimation->getAnimChannel(pBone->getJointId());
        if (0 != pAnimChannel)
        {
            glm::vec4* pRotate = pAnimChannel->getTimeRorate(time);
            pBone->setRotate(pRotate);
        }
    }
    return;
}

void RenderScene::updateBoneMatrix(Bone* pBone, glm::mat4& matrix)
{
    if (0 == pBone)
    {
        return;
    }

    if (pBone->getJointFlag())
    {
        updateJointInfo(pBone);
        glm::mat4 transMatrix = glm::translate(pBone->getTranslate());
        glm::mat4 scaleMatrix = glm::scale(pBone->getScale());
        glm::vec4 rotate = pBone->getRotate();
        float x2 = rotate.x * rotate.x;
        float y2 = rotate.y * rotate.y;
        float z2 = rotate.z * rotate.z;
        float xy = rotate.x * rotate.y;
        float xz = rotate.x * rotate.z;
        float yz = rotate.y * rotate.z;
        //float cosw = cos(rotate.w * 3.14159f / 180.0f);
        //float sinw = sin(rotate.w * 3.14159f / 180.0f);
        float cosw = cos(-rotate.w);
        float sinw = sin(-rotate.w);
        glm::mat4 rotateMatrix = glm::mat4(
            glm::vec4(x2*(1-cosw)+cosw, xy*(1-cosw)-rotate.x*sinw, xz*(1-cosw)+rotate.y*sinw, 0),
            glm::vec4(xy*(1-cosw)+rotate.x*sinw, y2*(1-cosw)+cosw, yz*(1-cosw)+rotate.z*sinw, 0),
            glm::vec4(xy*(1-cosw)-rotate.y*sinw, xy*(1-cosw)+rotate.x*sinw, z2*(1-cosw)+cosw, 0),
            glm::vec4(0.0f, 0.0f, 0.0f, 1.0f));
        glm::mat4 localMatrix = transMatrix * rotateMatrix * scaleMatrix;
        pBone->setLocalMatrix(&localMatrix);
    }

    glm::mat4 globleMatrix;
    globleMatrix = matrix * pBone->getLocalMatrix();
    pBone->setGlobleMatrix(&globleMatrix);

    if (pBone->hasChildren())
    {
        unsigned int count = pBone->getChildBoneSize();
        for (unsigned int i = 0; i < count; ++i)
        {
            updateBoneMatrix(pBone->getChildBone(i), globleMatrix);
        }
    }

    return;
}

bool RenderScene::setStates(Technique *pTechnique)
{
    if (1 == pTechnique->getTechState()->blendEnable)
    {
        glEnable(GL_BLEND);
        glBlendEquation(pTechnique->getTechState()->blendEquation);
        glBlendFunc(pTechnique->getTechState()->blendFuncSfactor,
            pTechnique->getTechState()->blendFuncDfactor);
    }
    else
        glDisable(GL_BLEND);

    if (1 == pTechnique->getTechState()->cullFaceEnable)
        glEnable(GL_CULL_FACE);
    else
        glDisable(GL_CULL_FACE);

    if (1 == pTechnique->getTechState()->depthMask)
        glDepthMask(GL_TRUE);
    else
        glDepthMask(GL_FALSE);

    if (1 == pTechnique->getTechState()->depthTestEnable)
        glEnable(GL_DEPTH_TEST);
    else
        glDisable(GL_DEPTH_TEST);

    return true;
}

CPhysicalCamera* RenderScene::getCamera()
{
    return sCamera;
}

void RenderScene::renderToBuffer(glTFViewport* pViewport, unsigned char * buffer)
{
    createRenderObj(pViewport->width, pViewport->height);
    createTextureObj(pViewport->width, pViewport->height);
    createFrameBufferObj(pViewport->width, pViewport->height);
    glBindFramebuffer(GL_FRAMEBUFFER, mFBOID);
    glViewport(pViewport->x, pViewport->y, pViewport->width, pViewport->height);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
    unsigned int pSize = getPrimitiveVecSize();
    for (unsigned int i = 0; i < pSize; ++i)
        renderPrimitive(getPrimitiveVec(i));
    glDisable(GL_DEPTH_TEST);
    bool result = true;
    GLenum fbResult = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if( fbResult != GL_FRAMEBUFFER_COMPLETE )
        result = false;
    glReadPixels(0, 0, pViewport->width, pViewport->height, GL_BGR, 
        GL_UNSIGNED_BYTE, buffer);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glDeleteFramebuffers(1, &mFBOID);
    glDeleteRenderbuffers(1,&mRBOID);
    glDeleteTextures(1,&mTextureObj);
}

void RenderScene::createRenderObj(int width, int height)
{
    glGenRenderbuffers(1, &mRBOID);
    glBindRenderbuffer(GL_RENDERBUFFER, mRBOID);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height);
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
}

void RenderScene::createTextureObj(int width, int height)
{
    glGenTextures(1, &mTextureObj);
    glBindTexture(GL_TEXTURE_2D, mTextureObj);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, 
        GL_UNSIGNED_BYTE, NULL);
    glBindTexture(GL_TEXTURE_2D, 0);
}

bool RenderScene::createFrameBufferObj(int width, int height)
{
    GLenum status;
    glGenFramebuffers(1, &mFBOID);
    status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (status != GL_FRAMEBUFFER_COMPLETE)
        return false;
    glBindFramebuffer(GL_FRAMEBUFFER, mFBOID);
    glBindTexture(GL_TEXTURE_2D, mTextureObj);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, 
        mTextureObj, 0);
    status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (status != GL_FRAMEBUFFER_COMPLETE)
        return false;
    glBindTexture(GL_TEXTURE_2D, 0);
    glBindRenderbuffer(GL_RENDERBUFFER, mRBOID);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, 
        GL_RENDERBUFFER, mRBOID);
    status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (status != GL_FRAMEBUFFER_COMPLETE)
        return false;
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    return true;
}

void RenderScene::getCameraPos(glm::vec3* Eye,glm::vec3* view,glm::vec3* up)
{
    *Eye =  sCamera->vEye;
    *view = sCamera->vView;
    *up = sCamera->vUp;
}

glm::vec3 RenderScene::getModelCenterPos()
{
    glm::vec3 vDelta = pScene->getVertexMax() - pScene->getVertexMin();
    glm::vec3 vHalf =  pScene->getVertexMin() + (vDelta / 2.0f);
    return vHalf;
}

void RenderScene::renderRotateCamera(double horizontal, double vertical,
    double planar, double time)
{
    sCamera->rotateCamera(horizontal,vertical,planar,time);
}

void RenderScene::renderMoveCamera(double x, double y, double z,
    double time)
{
    sCamera->moveCamera(x,y,z,time);
}

double RenderScene::getModelSize()
{
    return sCamera->flength;
}

void RenderScene::startAnimation()
{
    this->mAnimationPlay = true;
}

void RenderScene::stopAnimation()
{
    this->mAnimationPlay = false;
}

void RenderScene::setAnimTime(double time)
{
    this->mLastPlaying = clock() / double(CLOCKS_PER_SEC) - time;
}

double RenderScene::getAnimTime()
{
    return fmod(this->mCurrentTime, this->mDuration);
}

void RenderScene::setAnimLoop(int loop)
{
    if (0 != loop)
        this->mAnimationLoop = true;
    else
        this->mAnimationLoop = false;
}

int  RenderScene::getAnimLoop()
{
    return this->mAnimationLoop ? 1 : 0;
}

double RenderScene::getAnimDuration()
{
    return this->mDuration;
}

int  RenderScene::isAnimPlay()
{
    return this->mAnimationPlay ?  1 : 0;
}