/*
 * This file is part of the LibreOffice project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#ifndef RENDERSCENE_H
#define RENDERSCENE_H

#include "Shaders.h"
#include "Common.h"
#include "Texture.h"
#include "Camera.h"
#include "LoadScene.h"
#include "FPSCounter.h"

using namespace glTF;

static CPhysicalCamera* sCamera = NULL;

class RenderPrimitive
{
public:
    class Material* getMaterial();
    void setMaterial(class Material* pMaterial);

    unsigned int getVerterCount();
    void setVerterCount(unsigned int count);

    unsigned int getIndicesCount();
    void setIndicesCount(unsigned int count);

    int getShaderVertexId();
    void setShaderVertexId(int id);

    int getShaderNormalId();
    void setShaderNormalId(int id);

    int getShaderCoord0Id();
    void setShaderCoord0Id(int id);

    int getShaderWeightId();
    void setShaderWeightId(int id);

    int getShaderJointId();
    void setShaderJointId(int id);

    unsigned int getVertexBuffer();
    void setVertexBuffer(unsigned int bufferId);

    unsigned int getNormalBuffer();
    void setNormalBuffer(unsigned int bufferId);

    unsigned int getTexCoordBuffer();
    void setTexCoordBuffer(unsigned int bufferId);

    unsigned int getJointBuffer();
    void setJointBuffer(unsigned int bufferId);

    unsigned int getWeightBuffer();
    void setWeightBuffer(unsigned int bufferId);

    unsigned int getIndicesBuffer();
    void setIndicesBuffer(unsigned int bufferId);

    DataType_E getIndicesDataType();
    void setIndicesDataType(DataType_E type);

    void setProject(glm::mat4* pProject);
    glm::mat4* getProject();

    void setColor(glm::vec3* pColor);
    glm::vec3* getColor();

    void setMatrix(glm::mat4* pMatrix);
    glm::mat4* getMatrix();
    
    RenderPrimitive();
    ~RenderPrimitive();
private:
    glm::mat4* pMatrix;
    glm::mat4* pProject;
    glm::vec3* pColor;
    class Material* pMaterial;
    DataType_E mIndicesDataType;

    unsigned int mVerterCount;
    unsigned int mIndicesCount;

    int mShaderVertexId;
    int mShaderNormalId;
    int mShaderCoord0Id;
    int mShaderJointId;
    int mShaderWeightId;

    unsigned int mVertexBuffer;
    unsigned int mNormalBuffer;
    unsigned int mTexCoordBuffer;
    unsigned int mJointBuffer;
    unsigned int mWeightBuffer;

    unsigned int mIndicesBuffer;
};;

class RenderScene
{
public:
    // For camera
    glm::vec3 getModelCenterPos();
    void getCameraPos(glm::vec3* Eye,glm::vec3* view,glm::vec3* up);
    double getModelSize();
    CPhysicalCamera* getCamera();
    void renderRotateCamera(double horizontal, double vertical,
        double planar, double time);
    void renderMoveCamera(double x, double y, double z, double time);

    // For render scene
    glTFHandle* initScene(const glTFFile* jsonfile);
    void initRender(glTFHandle* handle);
    void prepareRender(void* lpParam);
    void render();
    void completeRender();
    void releaseRender(void* lpParam);

    // For bitmap
    void renderToBuffer(glTFViewport* pViewpoit, unsigned char * buffer);
    
    // For animation
    void startAnimation();
    void stopAnimation(); 
    void setAnimTime(double time);
    double getAnimTime();
    void setAnimLoop(int loop);
    int  getAnimLoop();
    double getAnimDuration();
    int  isAnimPlay();

    RenderScene();
    ~RenderScene();
private:
    glTFViewport viewport;
    Parser loadJson;
    CPhysicalCamera* pCamera;
    Light* pLight;
    glm::mat4* light0Transfrom;

    bool mAnimationPlay;
    bool mAnimationLoop;
    double mCurrentTime;
    double mLastPlaying;
    double mDuration;

    void initOpengl();
    class Scene* loadScene(const std::string& directory, const std::string& file);
    class Scene* loadScene(glTFHandle* handle);
    void createDefaultLight();
    void createDefaultCamera();
    void constructMesh(const std::string meshName, glm::mat4* pmatrix);
    void constructPrimitive(const Primitives* pPrim, glm::mat4* pmatrix);
    void setPrimitiveVec(class RenderPrimitive* pPrim);
    class RenderPrimitive* getPrimitiveVec(unsigned int index);
    unsigned int getPrimitiveVecSize();
    unsigned int bindAttribute(const class Attribute* pAttr);
    unsigned int bindIndices(const class Attribute* pAttr);
    void getAttributeShaderId(class Material* pMaterial, class RenderPrimitive* pRP);
    void bindAttributeBuffer(const Primitives* pPrimitive, class RenderPrimitive* pRP);
    void upLoadAttribute(unsigned int shaderId, unsigned int bufferId, int size);
    void drawTriangle(class RenderPrimitive* pPrimitive);
    void renderPrimitive(class RenderPrimitive* pPrimitive);

    void updateJointInfo(Bone* pBone);
    void updateBoneMatrix(Bone* pBone, glm::mat4& matrix);

    glm::mat4* calcProjectionMatrix();
    glm::vec3* calcLightColor();
    bool loadFPSShader();
    bool setStates(Technique* pTechnique);
    int getFPS();
    void printFPS( glTFViewport* pViewpoit );
    void createTextureObj(int width, int height);
    void createRenderObj(int width, int height);
    bool createFrameBufferObj(int width, int height);
    std::vector<class RenderPrimitive*> mPrimitiveVec;
    class ShaderProgram mShaderProgram;
    class Scene* pScene;
    GLuint mFBOID;
    GLuint mTextureObj;
    GLuint mRBOID;
#if ENABLE_FPS
    FPSCounter *pFPSCounter;
#endif
};

#endif
